# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (C) 2016 - David Zahn

#TODO
# how to handle various light PNGs?  settings per car?
# ini entries for all the tuned/upgraded options (Standard vs S3 for 458 Italia)
# make sure the maxGear is set to 6 by default

import os
import sys
import configparser
import traceback
import ac
import acsys
import datetime
import random
from time import clock, localtime, strftime
import subprocess

import platform
if platform.architecture()[0] == "64bit":
    sysdir=os.path.dirname(__file__)+'/stdlib64'
else:
    sysdir=os.path.dirname(__file__)+'/stdlib'
sys.path.insert(0, sysdir)
os.environ['PATH'] = os.environ['PATH'] + ";."

#from sim_info import info
import sim_info

#SETTING NOW ADJUSTED IN THE INI, CHANGING VALUES HERE WILL NOT AFFECT APP WHEN IT RUNS
Minimum = 5500.0 #set this the same as Warning to disable green light
Warning = 7100.0 #set this the same as Shift to disable a yellow warning light
Shift = 7300.0   #set this value to the RPM where you want a red light
maxGear = 6      #string value, normally 6
lightSize = 3    #may be 1, 2, 3, or 4
showGear = 1     #add this to the INI
usePNGs = 1      #make this more than binary?
showMin=1        #show the green light for exceeding minimum RPM?
showWarn=1       #show the warning light for exceeding warning RPM?
showTopGear=1    #show the lights even if the maxGear is reached
customFont = ""  # "Wednesday" font is provided as a "digital" font in the ZIP, but you can download your own
customFontSizeMultiplier = 1.0 #for use with custom fonts
customFontOffset = 0           #for use with custom fonts
customFontOffsetX = 0           #for use with custom fonts
customAppWidth = 0             #for use with PNGs
customAppHeight = 0            #for use with PNGs
randomRange = 500              #range to use when randomizing the RPM shift point
learningMode = 1               #allow learning mode at the drag strip?
ignoreRecentTrack = 0          #set to 1 to disable saving/loading recent tracks
pngFolder = "apps/python/EsoticShiftLight/" #this is the default folder where PNGs are located, can be overridden on a per car basis in the INI
#END OF USER ADJUSTABLE SETTINGS

#CHANGING THESE VALUES MAY RESULT IN UNINTENDED CONSEQUENCES
app = 0
dic = {}
fontSize = 300
gearFontSize = 20
Offset = -282
gearOffset = 10
app_size_x = 60
app_size_y = 60
background = 0.7
shiftLabel = 0
gearLabel = 0
track = None
config = None
configFile = "apps/python/EsoticShiftLight/settings.ini"
calibrate = 1
LapTotal = 0
PrevLap = 0
RecentRPM = 0
RecentGear = "N"
RecentTrackAndConfig = ""
bNoneSaved = 1

#track previous values and only update screen when they change, or very infrequently
strLastGear = ""
intCounter = 0
strLastTimestamp = ""

colors = {
	"white"      : [255, 255, 255],
	"yellow"     : [255, 255,   0],
	"green"      : [  0, 170,   0],
	"red"        : [255,   0,   0],
    "purple"     : [238, 130, 238],
    "orange"     : [255, 140,   0],
    "black"      : [  0,   0,   0],
    "grey"       : [140, 140, 140]
}

def acMain(ac_version):
    global app, shiftLabel, gearLabel, app_size_x, app_size_y

    try:

        getSettings()

        if (usePNGs == 1 or usePNGs == 2) and customAppWidth > 0 and customAppHeight > 0:
            app_size_x = customAppWidth
            app_size_y = customAppHeight
        
        app = ac.newApp("EsoticShiftLight")
        ac.setSize(app, app_size_x, app_size_y)
                
        ac.setTitle(app, "") 
        ac.setIconPosition(app, 0, -9000)
		
        ac.setBackgroundOpacity(app, background)
        
        shiftLabel = ac.addLabel(app,"SHIFT")
        ac.setFontSize(shiftLabel, fontSize)
        ac.setPosition(shiftLabel, (app_size_x / 2), Offset)
        ac.setFontAlignment(shiftLabel, "center")
        ac.setSize(shiftLabel, 0, 0)
        ac.setText(shiftLabel, ".")
        ac.setBackgroundOpacity(shiftLabel, 0.0)
        
        if showGear == 1:
            gearLabel = ac.addLabel(app,"GEAR")
            ac.setFontSize(gearLabel, gearFontSize)
            ac.setPosition(gearLabel, app_size_x / 2, gearOffset)
            ac.setFontAlignment(gearLabel, "center")
            ac.setSize(gearLabel, 0, 0)
            ac.setText(gearLabel, "N")
            ac.setBackgroundOpacity(gearLabel, 1.0)
            ac.setFontColor(gearLabel, *rgb(colors["black"], 1.0))
        
        if not customFont == "":
            ac.initFont(0, customFont, 0, 0)
            ac.setCustomFont(gearLabel, customFont, 0, 0)
            ac.setPosition(gearLabel, (app_size_x / 2) + customFontOffsetX, customFontOffset)
            ac.setFontSize(gearLabel, gearFontSize * customFontSizeMultiplier)
            
        if usePNGs == 1:
            ac.setText(shiftLabel, "")
            ac.setBackgroundOpacity(app, 0.0)
        elif usePNGs == 2:
            ac.setText(shiftLabel, "")
        else:
            ac.setText(shiftLabel, ".")

            
        return "EsoticShiftLight"
        
    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        ConsoleLog('Esotic Shift Light Main Error (logged to file)')
        #ac.console('Esotic Shift Light Main Error (logged to file)')
        ac.log(repr(traceback.format_exception(exc_type, exc_value, exc_traceback)))

def HitSpacebar():
    FNULL = open(os.devnull, 'w')    #use this if you want to suppress output to stdout from the subprocess
    args = "apps\\python\\EsoticShiftLight\\HitSpacebar.exe"
    subprocess.call(args, stdout=FNULL, stderr=FNULL, shell=False)

def writeNewShiftRPMToINI(Gear, ShiftRPM):
    global config

    if not config is None:
        ConsoleLog("Writing Gear %d Shift RPM %.1f" % (Gear, ShiftRPM))
        car = ac.getCarName(0)
        section = car
        if config.has_section(section):
            gearSetting = "Gear%dShiftRPM" % Gear
            gearRPMValue = "%.1f" % ShiftRPM
            config.set(section, gearSetting, gearRPMValue)
            writeConfigFile(config)
            
        #RecentTrackAndConfig
        section = car + "_" + RecentTrackAndConfig
        if config.has_section(section):
            gearSetting = "Gear%dShiftRPM" % Gear
            gearRPMValue = "%.1f" % ShiftRPM
            config.set(section, gearSetting, gearRPMValue)
            writeConfigFile(config)
            
    else:
        ConsoleLog("Config File is not found")

def writeNewMaxGear():
    global config

    if not config is None:
        ConsoleLog("Writing maxGear = %d" % maxGear)
        car = ac.getCarName(0)
        section = car
        if config.has_section(section):
            gearSetting = "maxGear"
            config.set(section, gearSetting, maxGear)
            writeConfigFile(config)
    else:
        ConsoleLog("Config File is not found")

def writeNewMaxRPM():
    global config

    if not config is None:
        maxRPM = "maxRPM"
        if maxRPM in dic:
            ConsoleLog("Writing maxRPM = %.1f" % dic[maxRPM])
            car = ac.getCarName(0)
            section = car
            if config.has_section(section):
                #rpmSetting = "maxRPM"
                config.set(section, maxRPM, "%.1f" % dic[maxRPM])
                writeConfigFile(config)
    else:
        ConsoleLog("Config File is not found")
        
def getSettings():
    global Warning, Shift, maxGear, Offset, fontSize, lightSize, gearFontSize, gearOffset, showGear, usePNGs, showMin, showWarn, showTopGear, customFont, customFontOffset, customFontSizeMultiplier, customAppHeight, customAppWidth, dic, Minimum, learningMode, RecentTrackAndConfig, ignoreRecentTrack, pngFolder, customFontOffsetX
    
    try:
        car = ac.getCarName(0)
        ConsoleLog("Shift Light Car = %s" % car)
        config = getConfigFile()
        Track = ac.getTrackName(0)
        TrackConfig = ac.getTrackConfiguration(0)

        section = "settings"
        if not config.has_section(section):
            config.add_section(section)
            config.set(section, "lightSize", lightSize)
            config.set(section, "showGear", showGear)
            writeConfigFile(config)
        else:
            #showGear
            if config.has_option(section, "showGear"):
                showGear = config.getint(section, "showGear")
            else:
                config.set(section, "showGear", showGear)    
                writeConfigFile(config)
            
            #usePNGs    
            if config.has_option(section, "usePNGs"):
                usePNGs = config.getint(section, "usePNGs")
            else:
                config.set(section, "usePNGs", usePNGs)    
                writeConfigFile(config)

            #pngFolder
            if config.has_option(section, "pngFolder"):
                pngFolder = config.get(section, "pngFolder")
            else:
                config.set(section, "pngFolder", pngFolder)    
                writeConfigFile(config)
            
                
            #showMin
            if config.has_option(section, "showMin"):
                showMin = config.getint(section, "showMin")
            else:
                config.set(section, "showMin", showMin)    
                writeConfigFile(config)
            #showWarn
            if config.has_option(section, "showWarn"):
                showWarn = config.getint(section, "showWarn")
            else:
                config.set(section, "showWarn", showWarn)    
                writeConfigFile(config)
            #showTopGear
            if config.has_option(section, "showTopGear"):
                showTopGear = config.getint(section, "showTopGear")
            else:
                config.set(section, "showTopGear", showTopGear)    
                writeConfigFile(config)
            
            #customFont = "Wednesday" # "Wednesday" font is provided as a "digital" font in the ZIP, but you can download your own
            if config.has_option(section, "customFont"):
                customFont = config.get(section, "customFont")
            else:
                config.set(section, "customFont", customFont)    
                writeConfigFile(config)
            #customFontSizeMultiplier = 1.45 #for use with custom fonts
            if config.has_option(section, "customFontSizeMultiplier"):
                customFontSizeMultiplier = config.getfloat(section, "customFontSizeMultiplier")
            else:
                config.set(section, "customFontSizeMultiplier", customFontSizeMultiplier)    
                writeConfigFile(config)
            #customFontOffset = 0            #for use with custom fonts
            if config.has_option(section, "customFontOffset"):
                customFontOffset = config.getfloat(section, "customFontOffset")
            else:
                config.set(section, "customFontOffset", customFontOffset)    
                writeConfigFile(config)
            #customFontOffsetX = 0
            if config.has_option(section, "customFontOffsetX"):
                customFontOffsetX = config.getfloat(section, "customFontOffsetX")
            else:
                config.set(section, "customFontOffsetX", customFontOffsetX)    
                writeConfigFile(config)
            #customAppWidth = 40             #for use with PNGs
            if config.has_option(section, "customAppWidth"):
                customAppWidth = config.getint(section, "customAppWidth")
            else:
                config.set(section, "customAppWidth", customAppWidth)    
                writeConfigFile(config)
            #customAppHeight = 40            #for use with PNGs
            if config.has_option(section, "customAppHeight"):
                customAppHeight = config.getint(section, "customAppHeight")
            else:
                config.set(section, "customAppHeight", customAppHeight)    
                writeConfigFile(config)
            
            #learningMode
            if config.has_option(section, "learningMode"):
                learningMode = config.getint(section, "learningMode")
            else:
                config.set(section, "learningMode", learningMode)    
                writeConfigFile(config)

            #ignoreRecentTrack
            if config.has_option(section, "ignoreRecentTrack"):
                ignoreRecentTrack = config.getint(section, "ignoreRecentTrack")
            else:
                config.set(section, "ignoreRecentTrack", ignoreRecentTrack)    
                writeConfigFile(config)
                
                
            #if recent track not loaded then no worries
            if ignoreRecentTrack == 0:
                #RecentTrackAndConfig
                if not "drag" in Track:
                    #write to the INI
                    RecentTrackAndConfig = Track + "_" + TrackConfig
                    config.set(section, "RecentTrackAndConfig", RecentTrackAndConfig)    
                    writeConfigFile(config)
                    ConsoleLog("Saved RecentTrackAndConfig = %s" % RecentTrackAndConfig)
                else:
                    #load the setting
                    if config.has_option(section, "RecentTrackAndConfig"):
                        RecentTrackAndConfig = config.get(section, "RecentTrackAndConfig")
                        ConsoleLog("Loaded RecentTrackAndConfig = %s" % RecentTrackAndConfig)
            
            if config.has_option(section, "lightSize"):
                lightSize = config.getint(section, "lightSize")
                if lightSize == 1:
                    fontSize = 100
                    gearFontSize = 11
                    gearOffset = 21.6
                    Offset = -74
                elif lightSize == 2:
                    fontSize = 200
                    gearFontSize = 23
                    gearOffset = 12
                    Offset = -178
                elif lightSize == 3:
                    fontSize = 300
                    gearFontSize = 35
                    gearOffset = 4
                    Offset = -282
                elif lightSize == 4:
                    fontSize = 400
                    gearFontSize = 46
                    gearOffset = -4
                    Offset = -385
            
        #default entry for the car
        section = car
        if not config.has_section(section):
            config.add_section(section)
            config.set(section, "MinRPM", Minimum)
            config.set(section, "WarnRPM", Warning)
            config.set(section, "ShiftRPM", Shift)
            config.set(section, "maxGear", maxGear)
            writeConfigFile(config)
        else:
            #Minimum
            if config.has_option(section, "MinRPM"):
                Minimum = config.getfloat(section, "MinRPM")
            else:
                config.set(section, "MinRPM", Minimum)
                writeConfigFile(config)
                
            if config.has_option(section, "WarnRPM"):
                Warning = config.getfloat(section, "WarnRPM")
            if config.has_option(section, "ShiftRPM"):
                Shift = config.getfloat(section, "ShiftRPM")
            if config.has_option(section, "MaxGear"):
                maxGear = config.getint(section, "MaxGear")

            maxRPM = "maxRPM"
            if config.has_option(section, maxRPM):
                dic[maxRPM] = config.getfloat(section, "maxRPM")
                ConsoleLog("maxRPM = %.1f" % dic[maxRPM])

            #MAKE SURE THIS SUPPORT ALL SETTINGS PER CAR
            if config.has_option(section, "lightSize"):
                lightSize = config.getint(section, "lightSize")
                if lightSize == 1:
                    fontSize = 100
                    gearFontSize = 11
                    gearOffset = 21.6
                    Offset = -74
                elif lightSize == 2:
                    fontSize = 200
                    gearFontSize = 23
                    gearOffset = 12
                    Offset = -178
                elif lightSize == 3:
                    fontSize = 300
                    gearFontSize = 35
                    gearOffset = 4
                    Offset = -282
                elif lightSize == 4:
                    fontSize = 400
                    gearFontSize = 46
                    gearOffset = -4
                    Offset = -385
            
            #showMin
            if config.has_option(section, "showMin"):
                showMin = config.getint(section, "showMin")

            #showWarn
            if config.has_option(section, "showWarn"):
                showWarn = config.getint(section, "showWarn")            
            
            #showGear
            if config.has_option(section, "showGear"):
                showGear = config.getint(section, "showGear")

            #showTopGear
            if config.has_option(section, "showTopGear"):
                showTopGear = config.getint(section, "showTopGear")
                
            #usePNGs    
            if config.has_option(section, "usePNGs"):
                usePNGs = config.getint(section, "usePNGs")
                
            #pngFolder
            if config.has_option(section, "pngFolder"):
                pngFolder = config.get(section, "pngFolder")                
                
            if config.has_option(section, "customAppWidth"):
                customAppWidth = config.getint(section, "customAppWidth")
            
            if config.has_option(section, "customAppHeight"):
                customAppHeight = config.getint(section, "customAppHeight")

            #customFont
            if config.has_option(section, "customFont"):
                customFont = config.get(section, "customFont")                                
            #customFontSizeMultiplier = 1.45 #for use with custom fonts
            if config.has_option(section, "customFontSizeMultiplier"):
                customFontSizeMultiplier = config.getfloat(section, "customFontSizeMultiplier")
                
            if config.has_option(section, "customFontOffset"):
                customFontOffset = config.getfloat(section, "customFontOffset")
            if config.has_option(section, "customFontOffsetX"):
                customFontOffsetX = config.getfloat(section, "customFontOffsetX")
                
            #gear1shiftrpm
            for item in range(1,maxGear):
                gearShiftINI = "Gear%dShiftRPM" % item
                gearShiftDIC = "Gear%dShift" % item
                if config.has_option(section, gearShiftINI):
                    dic[gearShiftDIC] = config.getfloat(section, gearShiftINI)
                    ConsoleLog("Config Gear %d ShiftRPM = %.1f" % (item, dic[gearShiftDIC]))

        if not RecentTrackAndConfig == "":
            section = car + "_" + RecentTrackAndConfig
            if not config.has_section(section):
                config.add_section(section)
                config.set(section, "MinRPM", Minimum)
                config.set(section, "WarnRPM", Warning)
                config.set(section, "ShiftRPM", Shift)
                #config.set(section, "maxGear", maxGear) #this does not change per track
                
                for item in range(1,maxGear):
                    gearShiftINI = "Gear%dShiftRPM" % item
                    gearShiftDIC = "Gear%dShift" % item
                    if gearShiftDIC in dic:
                        config.set(section, gearShiftINI, dic[gearShiftDIC])

                writeConfigFile(config)
            else:
                #Minimum
                if config.has_option(section, "MinRPM"):
                    Minimum = config.getfloat(section, "MinRPM")
                else:
                    config.set(section, "MinRPM", Minimum)
                    writeConfigFile(config)
                    
                if config.has_option(section, "WarnRPM"):
                    Warning = config.getfloat(section, "WarnRPM")
                if config.has_option(section, "ShiftRPM"):
                    Shift = config.getfloat(section, "ShiftRPM")
                if config.has_option(section, "MaxGear"):
                    maxGear = config.getint(section, "MaxGear")

                #gear1shiftrpm
                for item in range(1,maxGear):
                    gearShiftINI = "Gear%dShiftRPM" % item
                    gearShiftDIC = "Gear%dShift" % item
                    if config.has_option(section, gearShiftINI):
                        dic[gearShiftDIC] = config.getfloat(section, gearShiftINI)
                        ConsoleLog("Config Gear %d ShiftRPM = %.1f" % (item, dic[gearShiftDIC]))

                    
        ConsoleLog("MinRPM = %.1f" % Minimum)        
        ConsoleLog("WarnRPM = %.1f" % Warning)
        ConsoleLog("ShiftRPM = %.1f" % Shift)
        ConsoleLog("maxGear = %d" % maxGear)
              
        return True
        #return section

    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        ConsoleLog('Esotic Shift Light Get Settings Error (logged to file)')
        #ac.console('Esotic Shift Light Get Settings Error (logged to file)')
        ac.log(repr(traceback.format_exception(exc_type, exc_value, exc_traceback)))        
       
def getConfigFile():
    global config

    if config is None:
        ConsoleLog("Esotic Shift Light loading config")
        config = configparser.RawConfigParser()
        config.read(configFile)
        
    return config
       
def writeConfigFile(config):

    with open(configFile, 'w') as configSave:
        configSave.write(';Generic settings ------------------------' + '\n')
        configSave.write(';lightSize can be 1, 2, 3, or 4.  Use this for a simple dot in predefined sizes.' + '\n')
        configSave.write(';Set showGear to 1 to display the gear inside the light.' + '\n')
        configSave.write(';usePNGs = 1 will use images for the lights instead of a simple dot.  Copy the relevant PNG files into the folder next to the EsoticShiftLight.py' + '\n')        
        configSave.write(';PNGFolder default is apps/python/EsoticShiftLight/.  You can set this value on a per car basis' + '\n')
        
        configSave.write(';Set showMin to 1 to show the green light, set it to 0 to not show it.' + '\n')
        configSave.write(';Set showWarn to 1 to show the yellow light, set it to 0 to not show it.' + '\n')
        configSave.write(';Set showTopGear to 1 to show the lights on the top gear, set to 0 to not show lights on the top gear.' + '\n')
        
        configSave.write(';Set customFont name of custom font to use for gearLabel.  Custom font must be in the content\\fonts folder.' + '\n')
        configSave.write(';Set customFontSizeMultiplier to a float value 1 to adjust the size of the custom font.' + '\n')
        configSave.write(';Set customFontOffset to a value to move the gear up (negative values) or down (positive values).' + '\n')
        configSave.write(';Set customAppWidth when using PNGs to adjust the width of the app and the images.' + '\n')
        configSave.write(';Set customAppHeight when using PNGs to adjust the height of the app and the images.' + '\n')
        configSave.write(';Set learningMode to 1 to enable autocalculating shift points at the drag strip.  Set to 0 to disable learning mode.' + '\n')
        configSave.write(';RecentTrackAndConfig is set automatically to help with setting autocalculate entries for track/config combos you are working on to prepare for a race.' + '\n')
        
        #ignorerecenttrack
        configSave.write(';ignoreRecentTrack disables creating entries and gearing for track driven before visiting drag strip.' + '\n')
        
        configSave.write(';' + '\n')
        configSave.write(';Vehicle Specific Settings ----------------------------.' + '\n')
        configSave.write(';Set MinRPM, WarnRPM and ShiftRPM to the RPM values that enables the green, yellow, and red lights.' + '\n')
        configSave.write(';Set maxGear to the highest gear in the car.' + '\n')

        config.write(configSave)
        
def rgb(color, a = 1, bg = False):
	r = color[0] / 255
	g = color[1] / 255
	b = color[2] / 255
	if bg == False:
		return r, g, b, a
	else:
		return r, g, b
	
def GearNumToString(Gear):

    strGear = "%d" % (Gear - 1)
    if Gear < 1:
        strGear = "R"
    elif Gear == 1:
        strGear = "N"
    
    return strGear

def GearNumAdjusted(Gear):

    Gear = (Gear - 1)
    
    if Gear < 0:
        Gear = 0
    
    return Gear

    
def LogToCSV(strValues, fileName):
    with open("EsoticLogs\\" + fileName + ".csv", "a") as myfile:
        myfile.write(strValues)
    
def ConsoleLog(message):
    #ac.console(message)
    ac.log(message)
    
def acUpdate(deltaT):
    global app, shiftLabel, gearLabel, track, dic, calibrate, LapTotal, PrevLap, RecentRPM, maxGear, bNoneSaved
    global strLastGear, intCounter, strLastTimestamp, RecentGear
    
    try:
                        
        sim_info_obj = sim_info.SimInfo()
        bSet = False

        strTimestamp = strftime("%I_%M_%S %p", localtime())
        if strTimestamp != strLastTimestamp:
            #SOME THINGS ONLY HAPPEN ONCE A SECOND, JUST IN CASE
            #ConsoleLog("Counter = %d when TimeStamp = %s" % (intCounter, strTimestamp))
            intCounter = 0
            strLastTimestamp = strTimestamp
            
            ac.drawBorder(app, 0)
            
            if usePNGs == 1:
                ac.setText(shiftLabel, "")
                ac.setBackgroundOpacity(app, 0.0)
            elif usePNGs == 2:
                ac.setBackgroundOpacity(app, 1.0)
                ac.setText(shiftLabel, "")
            else:
                #ac.setBackgroundOpacity(app, background)
                ac.setBackgroundOpacity(app, 0.0)
                ac.setText(shiftLabel, ".")

            RecentRPM = 0.0
                
        else:
            intCounter = intCounter + 1
            #return True
        
               
        #try this for live and replay? status == live/replay and session == practice or hotlap (but not race)
        #if (sim_info_obj.graphics.status == 2 or sim_info_obj.graphics.status == 1) and (sim_info_obj.graphics.session == 0 or sim_info_obj.graphics.session == 3):
        if (sim_info_obj.graphics.status == 2 or sim_info_obj.graphics.status == 1):
        
            RPM = ac.getCarState(0, acsys.CS.RPM)
            SpeedKPH = ac.getCarState(0, acsys.CS.SpeedKMH)
            strGear = GearNumToString(ac.getCarState(0, acsys.CS.Gear))
            numGear = GearNumAdjusted(ac.getCarState(0, acsys.CS.Gear))
            AG_X, AG_Y, AG_Z = ac.getCarState(0, acsys.CS.AccG)
            Gas = ac.getCarState(0,acsys.CS.Gas)
            normalizedSplinePosition = ac.getCarState(0, acsys.CS.NormalizedSplinePosition)
            Brake = ac.getCarState(0,acsys.CS.Brake)
            LapCount = ac.getCarState(0,acsys.CS.LapCount)

            if RecentGear != strGear:
                RecentGear = strGear
                RecentRPM = 0
     
            #some cars bounce off the rev limit and cause flickering
            #in the shift light. this code tries to mitigate that behavior
            if RPM > RecentRPM:
                RecentRPM = RPM
            elif RecentRPM > RPM:
                RPM = RecentRPM
            
            if numGear > maxGear:
                maxGear = numGear
                writeNewMaxGear()

            gearMaxRPM = "max%dRPM" % numGear #this is regardless of gear or calibration/otherwise?
            gearLastMaxRPM = "max%dRPM" % (numGear - 1)
            if not gearMaxRPM in dic:
                dic[gearMaxRPM] = RPM
            elif RPM > dic[gearMaxRPM]:
                dic[gearMaxRPM] = RPM

            maxRPM = "maxRPM" #this is regardless of gear or calibration/otherwise?
            if not maxRPM in dic:
                dic[maxRPM] = RPM
                writeNewMaxRPM()
            elif RPM > dic[maxRPM]:
                dic[maxRPM] = RPM
                writeNewMaxRPM()
                
            gearShift = "Gear%dShift" % (numGear)
            gearShiftRandom = "Gear%dShiftRandom" % (numGear)
            gearLastShift = "Gear%dShift" % (numGear - 1)
            gearLastShiftRandom = "Gear%dShiftRandom" % (numGear - 1)
            gearShifted = "gearShifted"
            shiftRPM = Shift
            
            if gearShiftRandom in dic:
                shiftRPM = dic[gearShiftRandom]
            elif gearShift in dic:
                shiftRPM = dic[gearShift]
            
            if False: # Gas > 0.98:
                #Gear, KPH, RPM, AccG
                logText = "%s, %.4f, %.4f, %.4f, %.4f, %.4f, %.4f \n" % (strGear, SpeedKPH, RPM, AG_X, AG_Y, AG_Z, normalizedSplinePosition)
                LogToCSV(logText, "TESTING")
                #ac.setText(shiftLabel, "KPH = %.3f\nRPM = %.3f" % (SpeedKPH, RPM))
        
            if showGear == 1:
                #ac.console("setting gear = %s" % Gear)
                if strGear != strLastGear:
                    ac.setText(gearLabel, "%s" % strGear)
                    strLastGear = strGear
                            
            if numGear != maxGear or (showTopGear == 1):    
                alpha = 1.0
                if RPM > shiftRPM:
                    #ac.console("SHIFT")
                    bSet = True
                    #only do this once per gear
                    #use this only with a keyboard when the spacebar is upshift
                    if numGear > 0 and False:
                        if gearShifted in dic:
                            if dic[gearShifted] != numGear:
                                HitSpacebar()
                                dic[gearShifted] = numGear
                        else:
                            HitSpacebar()
                            dic[gearShifted] = numGear
                        
                    if usePNGs == 1:
                        #message = "using a PNG at " + str(datetime.datetime.now())
                        #ConsoleLog(message)
                        ac.setBackgroundTexture(app, pngFolder + "RedLight.png")
                    elif usePNGs == 2:
                        #ac.setBackgroundColor(app, *rgb(colors["red"], alpha, True)) 
                        ac.setBackgroundColor(app, 1.0, 0.0, 0.0)
                    else:
                        ac.setFontColor(shiftLabel, *rgb(colors["red"], alpha))
                        ac.setBackgroundOpacity(app, 0.0)
                elif RPM > Warning and showWarn == 1:
                    #ac.console("WARN")
                    bSet = True
                    if usePNGs == 1:
                        ac.setBackgroundTexture(app, pngFolder + "YellowLight.png")
                    elif usePNGs == 2:
                        #ac.setBackgroundColor(app, *rgb(colors["yellow"], alpha, True)) 
                        ac.setBackgroundColor(app, 1.0, 1.0, 0.0)
                    else:
                        ac.setFontColor(shiftLabel, *rgb(colors["yellow"], alpha))
                        ac.setBackgroundOpacity(app, 0.0)
                elif RPM > Minimum and showMin == 1:
                    #ac.console("WARN")
                    bSet = True
                    if usePNGs == 1:
                        ac.setBackgroundTexture(app, pngFolder + "GreenLight.png")
                    elif usePNGs == 2:
                        ac.setBackgroundColor(app, *rgb(colors["green"], alpha, True)) 
                        ac.setBackgroundColor(app, 0.0, 1.0, 0.0)
                    else:
                        ac.setFontColor(shiftLabel, *rgb(colors["green"], alpha))
                        ac.setBackgroundOpacity(app, 0.0)

            if SpeedKPH > 2 and bSet == False:
                if showGear == 1:
                    if usePNGs == 1:
                        ac.setBackgroundTexture(app, pngFolder + "GreyLight.png")
                    elif usePNGs == 2:
                        #ac.setBackgroundColor(app, *rgb(colors["grey"], alpha, True)) 
                        ac.setBackgroundColor(app, 0.5, 0.5, 0.5)
                    else:                
                        ac.setFontColor(shiftLabel, *rgb(colors["grey"], 1.0))
                        ac.setBackgroundOpacity(app, 0.0)
                else:
                    if usePNGs == 1:
                        ac.setBackgroundTexture(app, pngFolder + "WhiteLight.png")
                    elif usePNGs == 2:
                        #ac.setBackgroundColor(app, *rgb(colors["grey"], alpha, True)) 
                        ac.setBackgroundColor(app, 0.5, 0.5, 0.5)
                    else:
                        #ac.console("RESET")
                        #ac.setFontColor(shiftLabel, *rgb(colors["white"], 0.0))
                        ac.setFontColor(shiftLabel, *rgb(colors["grey"], 1.0))
                        ac.setBackgroundOpacity(app, 0.0)
            if SpeedKPH < 2:
                if usePNGs == 1:
                    ac.setBackgroundTexture(app, pngFolder + "GreyLight.png")
                elif usePNGs == 2:
                    #ac.setBackgroundColor(app, *rgb(colors["white"], alpha, True)) 
                    #ac.setBackgroundColor(app, 1.0, 1.0, 1.0)
                    ac.setBackgroundOpacity(app, 1.0)
                    ac.setBackgroundColor(app, 0.5, 0.5, 0.5, 1.0)
                else:
                    ac.setFontColor(shiftLabel, *rgb(colors["white"], 1.0))

                    
            #only update the controls if need be
            
                    
                    
            if track is None:
                track = ac.getTrackName(0)
                ConsoleLog("Track = " + track)
    
            #if learningMode is allowed then attempt to auto calculate
            #optimum shift points at the drag strip
            if "drag" in track and learningMode == 1:
                message = "learning mode at " + str(datetime.datetime.now())
                #ac.console(message)
                
                if numGear == 1 or normalizedSplinePosition == 0.0:
                    bNoneSaved = 1
                
                if normalizedSplinePosition > 0.0:
                    if LapCount == 0:
                            PrevLap = 0                
                    #only increment this at the end of the track when braking
                    if Brake > 0.1:
                        if LapCount > PrevLap:
                            PrevLap = LapCount
                            LapTotal = LapTotal + 1
                            message = "LapTotal = %d" % LapTotal
                            ConsoleLog(message)
                    
                    if calibrate == 1:
                        message = "calibrating at " + str(datetime.datetime.now())
                        #ac.console(message)
                        gearStartKPHAttr = "Gear%dStartKPH" % numGear
                        gearLastStartKPHAttr = "Gear%dStartKPH" % (numGear - 1)
                        gearLastEndKPHAttr = "Gear%dEndKPH" % (numGear - 1) # we don't know this until we shift
                        gearLastMidKPH = "Gear%dMidKPH" % (numGear - 1)     # must wait for shift
                        #ac.console("100")
                        #set the variables for all the Gear
                        if numGear == 1:
                            #ac.console("200")
                            if not gearStartKPHAttr in dic:
                                #ac.console("setting gear %d speeds" % numGear)
                                #first gear always starts at 0 KPH
                                dic[gearStartKPHAttr] = 0
                        elif numGear > 1:
                            #ac.console("300")
                            if not gearLastEndKPHAttr in dic:
                                #ac.console("setting gear %d speeds" % numGear)
                                #ac.console("310")
                                dic[gearLastEndKPHAttr] = SpeedKPH
                                #ac.console("320")
                                dic[gearStartKPHAttr] = SpeedKPH
                                #ac.console("330")
                                dic[gearLastMidKPH] = dic[gearLastStartKPHAttr] + ((dic[gearLastEndKPHAttr] - dic[gearLastStartKPHAttr]) / 2.0)
                                
                                message = "Gear %d StartKPH = %d, Mid KPH = %d, End KPH = %d" % ((numGear - 1), dic[gearLastStartKPHAttr], dic[gearLastMidKPH], dic[gearLastEndKPHAttr])
                                ConsoleLog(message)
                                
                                if gearLastMaxRPM in dic:
                                    message = "Gear %d maxRPM = %.1f" % (numGear -1, dic[gearLastMaxRPM])
                                    ConsoleLog(message)
                                
                            if numGear == maxGear:                                
                                    
                                #ac.console("400")
                                gearLastStartKPHAttr = "Gear%dStartKPH" % (numGear)
                                #ac.console("410")
                                gearLastEndKPHAttr = "Gear%dEndKPH" % (numGear)
                                #ac.console("420")
                                gearLastMidKPH = "Gear%dMidKPH" % (numGear)    
                                #ac.console("430")
                                if not gearLastEndKPHAttr in dic:
                                    #ac.console("setting gear %d speeds" % numGear)
                                    dic[gearLastEndKPHAttr] = SpeedKPH
                                else:
                                    if SpeedKPH > dic[gearLastEndKPHAttr]:
                                        dic[gearLastEndKPHAttr] = SpeedKPH
                                        dic[gearLastMidKPH] = dic[gearLastStartKPHAttr] + ((dic[gearLastEndKPHAttr] - dic[gearLastStartKPHAttr]) / 2.0)
                                        
                                if Brake > 0.1:
                                    ConsoleLog("end of calibration")
                                    if maxRPM in dic:
                                        ConsoleLog("maxRPM = %.1f" % dic[maxRPM])
                                    calibrate = 0
                                    message = "Gear %d StartKPH = %d, Mid KPH = %d, End KPH = %d" % (numGear, dic[gearLastStartKPHAttr], dic[gearLastMidKPH], dic[gearLastEndKPHAttr])
                                    ConsoleLog(message)

                    else:
                    
                        if Gas > 0.9 and numGear > 0:
                            #only track this under heavy acceleration
                            #track the gear and the PoT for the mid
                            #track best PoT len for the higher gear
                            gearMidKPH = "Gear%dMidKPH" % (numGear)
                            gearMidPoT = "Gear%dLap%dMidPoT" % (numGear, LapTotal)
                            gearLastMidPoT = "Gear%dLap%dMidPoT" % (numGear - 1, LapTotal)
                            
                            gearLastMaxRPM = "Gear%dLap%dMaxRPM" % (numGear - 1, LapTotal)
                            gearBestPoTLen = "Gear%dBestPoTLen" % (numGear)
                            
                            if not gearShift in dic:
                                #ConsoleLog("Setting Gear %d ShiftRPM = %d at %dKPH" % (numGear, shiftRPM, SpeedKPH))
                                dic[gearShift] = shiftRPM
                                                    
                            #not gearMidPoT in dic and 
                            if not gearMidPoT in dic and SpeedKPH > dic[gearMidKPH]:
                                dic[gearMidPoT] = normalizedSplinePosition
                                message = "Gear %d Mid PoT = %.4f" % (numGear, normalizedSplinePosition)
                                #ConsoleLog(message)
                                if gearLastMaxRPM in dic:
                                    message = "RPM Delay: Shift RPM = %d, ActualShift = %d, Difference = %d" % (shiftRPM, dic[gearLastMaxRPM], dic[gearLastMaxRPM] - shiftRPM)
                                    #ConsoleLog(message)
                                if numGear > 1:
                                    if gearLastMidPoT in dic:
                                        gearCalcedPoTLen = dic[gearMidPoT] - dic[gearLastMidPoT]
                                        if not gearBestPoTLen in dic:
                                            dic[gearBestPoTLen] = gearCalcedPoTLen
                                            message = "Initial Gear %d PoT Len = %.4f" % (numGear, gearCalcedPoTLen)
                                            ConsoleLog(message)
                                            if gearLastShift in dic:
                                                writeNewShiftRPMToINI(numGear - 1, dic[gearLastShift])
                                                bNoneSaved = 0
                                                #ConsoleLog("created gear %d shift point %.1f to INI" % (numGear - 1, dic[gearLastShift]))
                                        if gearCalcedPoTLen < dic[gearBestPoTLen]:
                                            dic[gearBestPoTLen] = gearCalcedPoTLen
                                            message = "Best Gear %d PoT Len = %.4f" % (numGear, gearCalcedPoTLen)
                                            ConsoleLog(message)
                                            #save new shiftRPM to the dic and the config
                                            if gearLastShiftRandom in dic:
                                                dic[gearLastShift] = dic[gearLastShiftRandom]
                                                writeNewShiftRPMToINI(numGear - 1, dic[gearLastShiftRandom])
                                                bNoneSaved = 0
                                                #ConsoleLog("saved gear %d shift point %.1f to INI" % (numGear - 1, dic[gearLastShift]))
                                    else:
                                        ConsoleLog("Gear %d MidPoT not found" % (numGear - 1))
                                        
                                #wiggle shift point for next time?
                                randomAmount = (((random.random() - 0.5) * 2) * randomRange)
                                #ConsoleLog("Gear %d Starting RPM = %.1f, randomAmount = %.1f" % (numGear, dic[gearShift], randomAmount))
                                dic[gearShiftRandom] = dic[gearShift] + randomAmount
                                if maxRPM in dic:
                                    if dic[gearShiftRandom] > dic[maxRPM] - 150 and dic[gearShiftRandom] > dic[gearShift]:
                                        #random value can't be higher than 100 RPM less than maxRPM this gear has been in
                                        dic[gearShiftRandom] = dic[maxRPM] - 150
                                message = "Gear %d ShiftRandom = %.4f" % (numGear, dic[gearShiftRandom])
                                #ConsoleLog(message)

                if (normalizedSplinePosition == 0.0 or normalizedSplinePosition > 0.81) and bNoneSaved == 1:
                    if usePNGs == 1:
                        ac.setBackgroundTexture(app, pngFolder + "PurpleLight.png")
                    else:                
                        ac.setFontColor(shiftLabel, *rgb(colors["purple"], 1.0))
                    
                #ac.console("learning mode")
                #learning mode
                #StartingSpeed = 20 #KPH #negating a bad launch?
                #if KPH > StartingSpeed:                
                #use the first run to create baseline data
                #then after that start tweaking
                #    DefaultRPM = Shift
                #    gearAttr = "Gear_" + Gear
                #    if gearAttr in dic:
                #        DefaultRPM = dic[gearAttr]
                #    else:
                #        dic[gearAttr] = DefaultRPM
                    
                
                
    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        ConsoleLog('Esotic Shift Light Update Error (logged to file)')
        #ac.console('Esotic Shift Light Update Error (logged to file)')
        ac.log(repr(traceback.format_exception(exc_type, exc_value, exc_traceback)))
    

